import json
import re
from transformers import AutoTokenizer
import time

import unidecode
from querent.common.types.ingested_table import IngestedTables
from querent.kg.ner_helperfunctions.fixed_predicate import FixedPredicateExtractor
from querent.common.types.ingested_images import IngestedImages
from querent.config.core.opensource_llm_config import Opensource_LLM_Config
from querent.core.transformers.relationship_extraction_llm import RelationExtractor
from querent.kg.rel_helperfunctions.contextual_predicate import process_data
from querent.kg.ner_helperfunctions.contextual_embeddings import EntityEmbeddingExtractor
from querent.kg.ner_helperfunctions.fixed_entities import FixedEntityExtractor
from querent.kg.ner_helperfunctions.ner_llm_transformer import NER_LLM
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from querent.common.types.file_buffer import FileBuffer
from querent.logging.logger import setup_logger
from querent.kg.querent_kg import QuerentKG
from querent.config.graph_config import GraphConfig
from querent.kg.ner_helperfunctions.attn_scores import EntityAttentionExtractor
from querent.kg.ner_helperfunctions.filter_triples import TripleFilter
from querent.config.core.llm_config import LLM_Config
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore

class BERTLLM(BaseEngine):
    def __init__(
        self,
        input_queue:QuerentQueue,
        config: LLM_Config,
        Embedding=None
    ):  
        self.logger = setup_logger(__name__, "BERTLLM")
        super().__init__(input_queue)
        self.skip_inferences=config.skip_inferences
        try:
            self.graph_config = GraphConfig(identifier=config.name)
            self.contextual_graph = QuerentKG(self.graph_config)
            self.semantic_graph = QuerentKG(self.graph_config)
            self.file_buffer = FileBuffer()
            self.ner_tokenizer = AutoTokenizer.from_pretrained(config.ner_model_name)
            self.ner_model = NER_LLM.load_model(config.ner_model_name, "NER")
            self.ner_llm_instance = NER_LLM(provided_tokenizer=self.ner_tokenizer, provided_model=self.ner_model)
            self.nlp_model = NER_LLM.set_nlp_model(config.spacy_model_path)
            self.nlp_model = NER_LLM.get_class_variable()
            if not Embedding:
                self.create_emb = EmbeddingStore()
            else:
                self.create_emb = Embedding
            if not self.skip_inferences:
                mock_config = Opensource_LLM_Config(qa_template=config.user_context,
                                                    model_type = config.rel_model_type,
                                                    model_path = config.rel_model_path,
                                                    grammar_file_path = config.grammar_file_path,
                                                    emb_model_name = config.emb_model_name,
                                                    spacy_model_path = config.spacy_model_path,
                                                    nltk_path = config.nltk_path
                                                    )
                self.semantic_extractor = RelationExtractor(mock_config,self.create_emb)
            self.attn_scores_instance = EntityAttentionExtractor(model=self.ner_model, tokenizer=self.ner_tokenizer)
            self.enable_filtering = config.enable_filtering
            self.filter_params = config.filter_params or {}
            self.triple_filter = None
            if self.enable_filtering:
                self.triple_filter = TripleFilter(**self.filter_params)
            self.sample_entities = config.sample_entities
            self.fixed_entities = config.fixed_entities
            if self.fixed_entities and not self.sample_entities:
                raise ValueError("If specific entities are provided, their types should also be provided.")
            if self.fixed_entities and self.sample_entities:
                self.entity_context_extractor = FixedEntityExtractor(fixed_entities=self.fixed_entities, entity_types=self.sample_entities,model = self.nlp_model)
            elif self.sample_entities:
                self.entity_context_extractor = FixedEntityExtractor(entity_types=self.sample_entities, model = self.nlp_model)
            else:
                self.entity_context_extractor = None
            self.fixed_relationships = config.fixed_relationships
            self.sample_relationships = config.sample_relationships
            if self.fixed_relationships and not self.sample_relationships:
                raise ValueError("If specific predicates are provided, their types should also be provided.")
            if self.fixed_relationships and self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(fixed_predicates=self.fixed_relationships, predicate_types=self.sample_relationships,model = self.nlp_model)
                self.predicate_json = self.predicate_context_extractor.construct_predicate_json(self.fixed_relationships, self.sample_relationships)
                self.predicate_json_emb = self.create_emb.generate_relationship_embeddings(self.predicate_json)
            elif self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(predicate_types=self.sample_relationships,model = self.nlp_model)
                self.predicate_json = self.predicate_context_extractor.construct_predicate_json(relationship_types=self.sample_relationships)
                self.predicate_json_emb = self.create_emb.generate_relationship_embeddings(self.predicate_json)
            else:
                self.predicate_context_extractor = None
            self.user_context = config.user_context
            self.isConfinedSearch = config.is_confined_search
        except Exception as e:
            self.logger.error("Error initializing BERT LLM Class", e)
            raise e
        
 

    def validate(self) -> bool:
        return self.ner_model is not None and self.ner_tokenizer is not None

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    @staticmethod
    def validate_ingested_images(data: IngestedImages) -> bool:
        if data.is_error():
            
            return False

        return True
    async def process_images(self, data: IngestedImages):
        content = ""
        doc_entity_pairs = []
        doc_entity_pairs_ocr = []
        entity_ocr = []
        results = []
        blob = data.image
        try:
            doc_source = data.doc_source
            if not BERTLLM.validate_ingested_images(data):
                self.set_termination_event()                                      
                return
            if data.ocr_text:
                ocr_text = ' '.join(data.ocr_text)
            if data.text:
                content = ' '.join(data.text)
            file = data.file
            ocr_content = ocr_text
            if ocr_content or content:
                (entity_ocr, doc_entity_pairs_ocr) = self.ner_llm_instance.get_entity_pairs(isConfinedSearch= self.isConfinedSearch, 
                                                                                                  content=ocr_content,
                                                                                                  fixed_entities=self.fixed_entities,
                                                                                                  sample_entities=self.sample_entities)
                if len(doc_entity_pairs_ocr) >= 1:
                    results = doc_entity_pairs_ocr
                elif len(doc_entity_pairs_ocr) == 0:
                    if content and len(entity_ocr) >=1:
                        if self.fixed_entities:
                            content = self.entity_context_extractor.find_entity_sentences(content)
                        (_, doc_entity_pairs) = self.ner_llm_instance.get_entity_pairs(isConfinedSearch= self.isConfinedSearch, 
                                                                                                  content=content,
                                                                                                  fixed_entities=self.fixed_entities,
                                                                                                  sample_entities=self.sample_entities)
                        if len(doc_entity_pairs) > 0 and len(entity_ocr) >=1:
                            results = [self.ner_llm_instance.filter_matching_entities(doc_entity_pairs, entity_ocr)]
                        elif len(doc_entity_pairs) > 0 and len(entity_ocr) == 0:
                            # results = doc_entity_pairs
                            pass
                    else:
                        return        
                if len(results) > 0:
                    doc_entity_pairs = self.ner_llm_instance.remove_duplicates(results)
                    filtered_triples = process_data(doc_entity_pairs, file)
                    if self.skip_inferences:
                        return filtered_triples, file, self.ner_llm_instance
                    else:
                        unique_id = str(hash(data.image))
                        for triple in filtered_triples:
                            if not self.termination_event.is_set():
                                updated_tuple = self.ner_llm_instance.final_ingested_images_tuples(triple, create_embeddings=self.create_emb)
                                graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(updated_tuple))
                                if graph_json:
                                    current_state = EventState(event_type=EventType.Graph, timestamp=time.time(), payload=graph_json, file=file, doc_source=doc_source, image_id=unique_id)
                                    await self.set_state(new_state=current_state)
                                vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(updated_tuple, blob))
                                if vector_json:
                                    current_state = EventState(event_type=EventType.Vector, timestamp=time.time(), payload=vector_json, file=file, doc_source=doc_source, image_id=unique_id)
                                    await self.set_state(new_state=current_state)
            else:
                return        
        except Exception as e:
            self.logger.debug(f"Invalid {self.__class__.__name__} configuration. Unable to process tokens. {e}")
    
    async def process_tables(self, data: IngestedTables):
        return super().process_tables(data)

    async def process_code(self, data: IngestedCode):
        return super().process_code(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True

    def count_entity_pairs(self, doc_entity_pairs):
        total_pairs = 0
        for inner_list in doc_entity_pairs:
            total_pairs += len(inner_list)
            
        return total_pairs
    
    def set_filter_params(self, **kwargs):
        self.filter_params = kwargs
        if self.triple_filter:
            self.triple_filter.set_params(**kwargs)
        else:
            self.triple_filter = TripleFilter(**kwargs)
    
    
    async def process_tokens(self, data: IngestedTokens):
        doc_entity_pairs = []
        try:
            doc_source = data.doc_source
            if not BERTLLM.validate_ingested_tokens(data):
                    self.set_termination_event()                                      
                    return
            if data.data:
                clean_text = ' '.join(data.data)
            else:
                clean_text = data.data
            if not data.is_token_stream : 
                file, content = self.file_buffer.add_chunk(
                data.get_file_path(), clean_text)
            else:
                content = clean_text
                file = data.get_file_path()
            if content:
                if self.fixed_entities:
                    content = self.entity_context_extractor.find_entity_sentences(content)
                if self.fixed_relationships:
                    content = self.predicate_context_extractor.find_predicate_sentences(content)
                (_, doc_entity_pairs) = self.ner_llm_instance.get_entity_pairs(isConfinedSearch= self.isConfinedSearch, 
                                                                                                  content=content,
                                                                                                  fixed_entities=self.fixed_entities,
                                                                                                  sample_entities=self.sample_entities)
            else:
                return
            if self.sample_entities:
                doc_entity_pairs = self.entity_context_extractor.process_entity_types(doc_entities=doc_entity_pairs)
            if any(doc_entity_pairs):
                doc_entity_pairs = self.ner_llm_instance.remove_duplicates(doc_entity_pairs)
                pairs_withattn = self.attn_scores_instance.extract_and_append_attention_weights(doc_entity_pairs)
                if self.enable_filtering == True and not self.entity_context_extractor and self.count_entity_pairs(pairs_withattn)>1 and not self.predicate_context_extractor:
                    self.entity_embedding_extractor = EntityEmbeddingExtractor(self.ner_model, self.ner_tokenizer)
                    pairs_withemb = self.entity_embedding_extractor.extract_and_append_entity_embeddings(pairs_withattn)
                else:
                    pairs_withemb = pairs_withattn
                pairs_with_predicates = process_data(pairs_withemb, file)
                if self.enable_filtering == True and not self.entity_context_extractor and self.count_entity_pairs(pairs_withattn)>1 and not self.predicate_context_extractor:
                    cluster_output = self.triple_filter.cluster_triples(pairs_with_predicates)
                    clustered_triples = cluster_output['filtered_triples']
                    cluster_labels = cluster_output['cluster_labels']
                    cluster_persistence = cluster_output['cluster_persistence']
                    if clustered_triples:
                        filtered_triples, reduction_count = self.triple_filter.filter_triples(clustered_triples)
                    else:
                        self.logger.debug(f"Filtering in {self.__class__.__name__} producing 0 entity pairs. Filtering Disabled. ")
                        filtered_triples = pairs_with_predicates
                else:
                    filtered_triples = pairs_with_predicates
                if not filtered_triples:
                    return
                elif not self.skip_inferences:
                    relationships = self.semantic_extractor.process_tokens(filtered_triples, fixed_entities=(len(self.sample_entities) >= 1))
                    if len(relationships) > 0:
                        if self.fixed_relationships and self.sample_relationships:
                            embedding_triples = self.create_emb.generate_embeddings(relationships, relationship_finder=True, generate_embeddings_with_fixed_relationship = True)
                        elif self.sample_relationships:
                            embedding_triples = self.create_emb.generate_embeddings(relationships, relationship_finder=True)
                        else:
                            embedding_triples = self.create_emb.generate_embeddings(relationships)
                        if self.sample_relationships:
                            embedding_triples = self.predicate_context_extractor.update_embedding_triples_with_similarity(self.predicate_json_emb, embedding_triples)
                        for triple in embedding_triples:
                            if not self.termination_event.is_set():
                                graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(triple))
                                if graph_json:
                                    current_state = EventState(event_type=EventType.Graph, timestamp=time.time(), payload=graph_json, file=file, doc_source=doc_source)
                                    await self.set_state(new_state=current_state)
                                vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(triple))
                                if vector_json:
                                    current_state = EventState(event_type=EventType.Vector, timestamp=time.time(), payload=vector_json, file=file, doc_source=doc_source)
                                    await self.set_state(new_state=current_state)
                            else:
                                return
                    else:
                        return
                else:
                    return filtered_triples, file
            else:
                return
        except Exception as e:
            self.logger.debug(f"Invalid {self.__class__.__name__} configuration. Unable to process tokens. {e}")
